from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import numpy as np 
import nibabel as nib
import pandas as pd
import json
import datetime
from pathlib import Path
import shlex
import subprocess
from nipype import logging as nlogging
from nilearn.maskers import NiftiLabelsMasker
from nilearn.maskers import NiftiMapsMasker
from nilearn import image

IFLOGGER=nlogging.getLogger('nipype.interface')

def roi_extract_proc(labels_dict,input_file,atlas_file,atlas_index, lesion_file):

    metadata_comments=""
    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    output_dir=cwd
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')

    lesion_inverse_img = None
    if lesion_file and not lesion_file == ".": 
        lesion_img = nib.load(lesion_file)
        lesion_data = lesion_img.get_fdata()
        inverse_data = np.zeros(lesion_data.shape)
        inverse_data[lesion_data < 1] = 1 
        lesion_inverse_img=nib.Nifti1Image(inverse_data, lesion_img.affine)

    
    if not session_label:
        roi_output_dir = os.path.join(cwd,f"sub-{participant_label}_roi_output_dir")
    else:
        roi_output_dir = os.path.join(cwd,f"sub-{participant_label}_ses-{session_label}_roi_output_dir")

    if not os.path.isdir(roi_output_dir):
        os.makedirs(roi_output_dir)

    if Path(atlas_file).suffix == ".mgz":
        mgzdir = os.path.join(cwd,'mgz_nii')
        if not os.path.isdir(mgzdir):
            os.makedirs(mgzdir)

        fs_command_base, fscontainer = getContainer(labels_dict,nodename="convMGZ2NII",SPECIFIC="FREESURFER_CONTAINER",LOGGER=IFLOGGER)
        atlas_file_nii = newfile(mgzdir,atlas_file,extension=".nii.gz")
        convMGZ2NII(atlas_file, atlas_file_nii, fs_command_base)
        atlas_file = atlas_file_nii

    if Path(input_file).suffix == ".mgz":
        mgzdir = os.path.join(cwd,'mgz_nii')
        if not os.path.isdir(mgzdir):
            os.makedirs(mgzdir)

        fs_command_base, fscontainer = getContainer(labels_dict,nodename="convMGZ2NII",SPECIFIC="FREESURFER_CONTAINER",LOGGER=IFLOGGER)
        input_file_nii = newfile(mgzdir,input_file,extension=".nii.gz")
        convMGZ2NII(input_file, input_file_nii, fs_command_base)
        input_file = input_file_nii

    if Path(lesion_file).suffix == ".mgz":
        mgzdir = os.path.join(cwd,'mgz_nii')
        if not os.path.isdir(mgzdir):
            os.makedirs(mgzdir)

        fs_command_base, fscontainer = getContainer(labels_dict,nodename="convMGZ2NII",SPECIFIC="FREESURFER_CONTAINER",LOGGER=IFLOGGER)
        lesion_file_nii = newfile(mgzdir,lesion_file,extension=".nii.gz")
        convMGZ2NII(lesion_file, lesion_file_nii, fs_command_base)
        lesion_file = lesion_file_nii

    atlas_type="3D"
    atlas_img  = nib.load(atlas_file)
    atlas_dim = 1
    atlas_shape = atlas_img.header.get_data_shape()
    if len(atlas_shape) > 3:
        if atlas_shape[3]> 1:
            atlas_type="4D"
            atlas_dim = atlas_shape[3]

    atlas_index_mode = None
    if getParams(labels_dict,'ATLAS_INDEX_MODE'):
        atlas_index_mode = getParams(labels_dict,'ATLAS_INDEX_MODE')
    elif getParams(labels_dict,'NEWATLAS_INDEX_MODE'):
        atlas_index_mode = getParams(labels_dict,'NEWATLAS_INDEX_MODE')

    if not atlas_index_mode:
        atlas_index_mode = "tsv"
    
    check_unknown_rois = False
    if getParams(labels_dict,'CHECK_UNKNOWN_ROIS'):
        check_unknown_rois = isTrue(getParams(labels_dict,'CHECK_UNKNOWN_ROIS'))

    if atlas_index.split(":")[0] == "get_freesurfer_atlas_index":
        lutfile = substitute_labels(atlas_index.split(":")[1],labels_dict)
        if "tsv" in atlas_index_mode:
            new_atlas_index=newfile(roi_output_dir,atlas_file,suffix="desc-index",extension=".tsv")
        else:
            new_atlas_index=newfile(roi_output_dir,atlas_file,suffix="desc-index",extension=".txt")
        atlas_index_json = newfile(roi_output_dir,new_atlas_index,extension="json")
        atlas_dict,atlas_index_out=get_freesurferatlas_index_mode(atlas_file,lutfile,new_atlas_index,atlas_index_mode=atlas_index_mode)
        atlas_index=new_atlas_index
        export_labels(atlas_dict,atlas_index_json)

    # See if atlas index is in right format
    labelfile_df=pd.read_csv(atlas_index,delim_whitespace=True)
    if "index" in labelfile_df.columns and "label" in labelfile_df:
        labels_index_list = labelfile_df["index"].tolist()
        labels_name_list = labelfile_df["label"].tolist()
    else:
        with open(atlas_index, 'r') as in_file:
            lines = in_file.readlines()
        labels_name_list = [x.replace('\n','') for x in lines]
        labels_index_list = range(1,len(labels_name_list)+1)

    measure_type = "3D"
    measure_img = nib.load(input_file)
    measure_data = measure_img.get_fdata()
    measure_shape = measure_img.header.get_data_shape()
    if len(measure_shape) > 3:
        if measure_shape[3] > 1:
            measure_type = "4D"

    if atlas_type == "4D":
        if lesion_inverse_img:
            NiftiMasker = NiftiMapsMasker(
                atlas_img,
                mask_img = lesion_inverse_img,
                Labels = labels_name_list,
                standardize=False
            )
        else:
            NiftiMasker = NiftiMapsMasker(
                atlas_img,
                Labels = labels_name_list,
                standardize=False
            )
    else:
        if lesion_inverse_img:
            NiftiMasker = NiftiLabelsMasker(
                atlas_img,
                mask_img = lesion_inverse_img,
                Labels = labels_name_list,
                standardize=False
            )
        else:
            NiftiMasker = NiftiLabelsMasker(
                atlas_img,
                Labels = labels_name_list,
                standardize=False
            )

    NiftiMasker.fit(input_file)
    signals = NiftiMasker.transform(input_file)
    num_rows = signals.shape[0]

    # check that rois exist:
    missing_rois = []
    unknown_rois=[]
    reconciled_labels = labels_name_list.copy()
    reconciled_signals = signals.copy()
    if atlas_type == "3D":
        atlas_data = atlas_img.get_fdata()
 
        for index in labels_index_list:
            lbl_index = labels_index_list.index(index)
            roi_sum = np.sum(atlas_data == int(index))
            if roi_sum == 0:               
                UTLOGGER.warn(f"WARNING: Roi Number {index} is missing from atlas {atlas_file}. Have you provided the correct labels in {atlas_index} ?")
                UTLOGGER.warn(f"WARNING: Roi Number {index} corresponds with ROI name : {labels_name_list[lbl_index]}")               
                if lbl_index not in missing_rois:
                    missing_rois.append(lbl_index)
                    if num_rows > 1:
                        insarr = np.array([np.nan for x in range(0,num_rows)])
                        reconciled_signals= np.insert(reconciled_signals, lbl_index,[insarr],axis=1)
                    else:
                        reconciled_signals = np.insert(reconciled_signals,lbl_index,np.nan)

                        
        for index in labels_index_list:
            lbl_index = labels_index_list.index(index)
            if measure_type == "3D":
                check = measure_data[atlas_data == int(index)]
            else:
                check = measure_data[atlas_data == int(index),:]
            if len(check) < 1:
                UTLOGGER.warn(f"WARNING: Roi Number {index} does not have any values in the measures file  {input_file}.")
                UTLOGGER.warn(f"WARNING: Roi Number {index} corresponds with ROI name : {labels_name_list[lbl_index]}")
                if lbl_index not in missing_rois:
                    missing_rois.append(lbl_index)
                    if num_rows > 1:
                        insarr = np.array([np.nan for x in range(0,num_rows)])
                        reconciled_signals= np.insert(reconciled_signals, lbl_index,[insarr],axis=1)
                    else:
                        reconciled_signals = np.insert(reconciled_signals,lbl_index,np.nan)

        # check for unknown rois, assume rous start from 1 and indexing starts from zero
        if check_unknown_rois:
            max_roi = int(np.max(atlas_data))
            for roi_avail in range(1,max_roi+1):
                if roi_avail not in labels_index_list:
                    UTLOGGER.warn(f"WARNING: The roi label; {roi_avail} is present in the atlas but missing in the the index {atlas_index}. Have you provided the correct atlas {atlas_file}")
                    if f"{roi_avail}_unknown" not in unknown_rois:
                        unknown_rois.append(f"{roi_avail}_unknown")
                        reconciled_labels.insert(roi_avail - 1,f"{roi_avail}_unknown")
                        if num_rows > 1:
                            insarr = np.array([np.nan for x in range(0,num_rows)])
                            reconciled_signals= np.insert(reconciled_signals, roi_avail - 1,[insarr],axis=1)
                        else:
                            reconciled_signals = np.insert(reconciled_signals,roi_avail - 1,np.nan)

    else:
        labels_len = len(labels_name_list)
        if not labels_len == atlas_dim:
            UTLOGGER.warn(f"WARNING: Mismatch between number of labels {len(labels_name_list)} and the size of 4D atlas {atlas_dim}. Please correct this.")
            if labels_len  < atlas_dim:
                if check_unknown_rois:
                    UTLOGGER.warn(f"WARNING: There is/are {atlas_dim - labels_len} unknown roi/s.")
                    for unk in range(1,atlas_dim - labels_len + 1):
                        unknown_rois.append(f"unknown{unk}")
                        reconciled_labels.append(f"unknown{unk}")
            else:
                UTLOGGER.warn(f"WARNING: There is/are {labels_len - atlas_dim} missing roi/s.")
                for miss in range(atlas_dim,labels_len):
                    missing_rois.append(miss)
                    if num_rows > 1:
                        insarr = np.array([np.nan for x in range(0,num_rows)])
                        reconciled_signals= np.insert(reconciled_signals, len(reconciled_signals),[insarr],axis=1)
                    else:
                        reconciled_signals = np.insert(reconciled_signals,len(reconciled_signals[0]),np.nan)

        atlas_index=0
        for img in image.iter_img(atlas_img):
            atlas_data = img.get_fdata()
            if measure_type == "3D":
                check = measure_data[atlas_data > 0]               
            else:
                check = measure_data[atlas_data > 0,:]
            if len(check) < 1:
                print(f"WARNING: Atlas Roi Volume {atlas_index} does not have any values in the measures file {input_file}.")
                missing_rois.insert(atlas_index,atlas_index+1)
                if num_rows > 1:
                    insarr = np.array([np.nan for x in range(0,num_rows)])
                    reconciled_signals= np.insert(reconciled_signals, atlas_index,[insarr],axis=1)
                else:
                    reconciled_signals = np.insert(reconciled_signals,atlas_index,np.nan)
    
            atlas_index = atlas_index + 1       

    if len(reconciled_signals.shape) > 1:
        df2=pd.DataFrame(reconciled_signals,columns=reconciled_labels)
    else:
        df2=pd.DataFrame([reconciled_signals],columns=reconciled_labels)
    
    csv_basename = ""

    atlas_name = getParams(labels_dict,'ATLAS_NAME')
    if not atlas_name:
        atlas_name = os.path.basename(atlas_file).split('.')[0].split('_')[0].split('-')[0]
    csv_basename = f"atlas-{atlas_name}"

    modality = get_bidstag("desc",input_file,True)
    if modality:
        modality=modality[-1].split(".")[0]
        csv_basename = csv_basename + "_" + modality
    else:
        modality = os.path.basename(input_file).split('.')[0].split('_')[-1].split('-')[-1]
        csv_basename = csv_basename + "_" + modality

    table_columns = [f"{atlas_name}.{x}.{modality}" for x in reconciled_labels]

    if not session_label:
        csv_basename  = f"sub-{participant_label}" + "_" + csv_basename
    else:
        csv_basename  = f"sub-{participant_label}_ses-{session_label}" + "_" + csv_basename

    roi_csv = os.path.join(roi_output_dir,'{}.csv'.format(csv_basename))

    if num_rows < 2:
        if session_label:
            df2.insert(0,"session_id",["ses-"+session_label])
            table_columns.insert(0,"session_id")
        else:
            df2.insert(0,"session_id",["NotProvided"])
            table_columns.insert(0,"session_id")

        df2.insert(0,"subject_id",["sub-"+participant_label])
        table_columns.insert(0,"subject_id")

        df2.columns = table_columns
        df2.to_csv(roi_csv,sep=",",header=True, index=False)
    else:
        flat_vals=[]
        flat_tablecolumns=[]
        for count in range(num_rows):
            flat_vals.extend(df2.iloc[count].to_list())
            flat_tablecolumns.extend([x + f".{count}" for x in table_columns])
        newdf=pd.DataFrame([flat_vals])
        if session_label:
            newdf.insert(0,"session_id",["ses-"+session_label])
            flat_tablecolumns.insert(0,"session_id")
        else:
            newdf.insert(0,"session_id",["NotProvided"])
            flat_tablecolumns.insert(0,"session_id")
        newdf.insert(0,"subject_id",["sub-"+participant_label])
        flat_tablecolumns.insert(0,"subject_id")
        newdf.columns = flat_tablecolumns
        newdf.to_csv(roi_csv,sep=",",header=True, index=False)


    metadata = {}
    roi_json = os.path.join(roi_output_dir,'{}.json'.format(csv_basename))
    metadata = updateParams(metadata,"Title","roi_extract")
    metadata = updateParams(metadata,"Description","Extract Measures from Image file using provided atlas.")
    metadata = updateParams(metadata,"DateCreated",datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S:%f"))
    metadata = updateParams(metadata,"Atlas File",atlas_file)
    metadata = updateParams(metadata,"Atlas Labels",atlas_index)
    metadata = updateParams(metadata,"Input File",input_file)
    metadata = updateParams(metadata,"Command","Nilearn NifiMasker")
    if metadata_comments:
        metadata = updateParams(metadata,"Comments",metadata_comments)

    export_labels(metadata,roi_json)


    out_files=[]
    out_files.insert(0,roi_csv)

    return {
        "roi_csv":roi_csv,
        "roi_output_dir":roi_output_dir,
        "output_dir":output_dir,
        "out_files":out_files
    }



class roi_extractInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    input_file = File(desc="input file")
    atlas_file = File(desc='atlas file')
    atlas_index = File(desc='atlas index file')
    lesion_file = File(desc='lesion file')

class roi_extractOutputSpec(TraitedSpec):
    roi_csv = File(desc='CSV file of results')
    roi_output_dir = traits.String(desc='roi output dir')
    output_dir = traits.String(desc='output dir')
    out_files = traits.List(desc='list of files')
    
class roi_extract_pan(BaseInterface):
    input_spec = roi_extractInputSpec
    output_spec = roi_extractOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = roi_extract_proc(
            self.inputs.labels_dict,
            self.inputs.input_file,
            self.inputs.atlas_file,
            self.inputs.atlas_index,
            self.inputs.lesion_file,
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="roi_extract_node",input_file="",atlas_file="",atlas_index="", lesion_file="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(roi_extract_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    pan_node.inputs.input_file = input_file
    pan_node.inputs.atlas_file =  atlas_file
    pan_node.inputs.atlas_index =  atlas_index
    pan_node.inputs.lesion_file =  lesion_file

    return pan_node





from panpipelines.utils.util_functions import *
from subprocess import run
import os
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

class panscript:

    def __init__(self,labels_dict,name='template_panscript',params="",command="",execution={}):

        # Create Node
        self.labels_dict = labels_dict
        self.name = name
        self.params = params
        self.command = command


    def pre_run(self):
        pass

    def run(self):

        self.pre_run()

        IFLOGGER.info(f"Running PAN script - {self.name}")
        command=self.command + " " + self.params

        pkgdir = os.path.abspath(os.path.dirname(__file__))
        IFLOGGER.info(f"Changing to panscript directory {pkgdir} to execute.")
        os.chdir(pkgdir)
        
        evaluated_command=substitute_labels(command, self.labels_dict)
        runCommand(evaluated_command,IFLOGGER)

        self.post_run()
        return self.get_results()

    def post_run(self):
        pass

    def get_results(self):
        self.results = {}
        return self.results








#!/usr/bin/env python
from panpipelines.utils.util_functions import *
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from functools import partial
import json
import pandas as pd
import os
import datetime
import logging
import sys
from panpipelines.version import __version__
from panpipelines import Factory

LOGGER = logger_setup("panpipelines.group_subjects", logging.DEBUG)
logger_addstdout(LOGGER, logging.INFO)

panFactory = Factory.getPANFactory()

def runGroupSubjects(participant_label, xnat_projects, session_label, pipeline, pipeline_class, pipeline_outdir, panpipe_labels,bids_dir,cred_user,cred_password, execution_json,analysis_level="group",panlabel=None):

    # get parent directory
    if not panlabel:
        panlabel=os.path.basename(os.path.dirname(pipeline_outdir))

    pipeline_outdir=os.path.join(pipeline_outdir,"group")
    if not os.path.exists(pipeline_outdir):
        os.makedirs(pipeline_outdir,exist_ok=True)
    
    LOGFILE=os.path.join(pipeline_outdir,f"group_{pipeline}_{panlabel}.log")
    logger_addfile(LOGGER, LOGFILE, logging.DEBUG)
    nipype_loggers_setup(logging.INFO,LOGFILE,logging.DEBUG)

    LOGGER.info(f"Pan Processing for group: Running {pipeline} pipeline")
    LOGGER.info(f"start logging to {LOGFILE}")

    panProcessor = panFactory.get_processflow(pipeline_class)

    PanProc = panProcessor(panpipe_labels,pipeline_outdir, participant_label, name=pipeline,LOGGER=LOGGER,execution=execution_json, analysis_level = analysis_level, participant_project=xnat_projects, participant_session=session_label)
    PanProc.run()

    LOGGER.debug(f"\nDump of configuration settings for group run of {pipeline}")
    LOGGER.debug("---------------------------------------------------------------------------------")
    LOGGER.debug(f"{panpipe_labels!r}")
    LOGGER.debug("\n------  environment dump -----------")
    LOGGER.debug(os.environ)

def parse_params():
    parser = ArgumentParser(description="Pan pipelines")
    PathExists = partial(path_exists, parser=parser)
    parser.add_argument("config_file", type=PathExists, help="Pipeline Configuration File")
    parser.add_argument("pipeline", help="group subjects running pipeline")
    parser.add_argument("credentials", help="credential file")
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))
    return parser

def main():
    
    parser=parse_params()
    args, unknown_args = parser.parse_known_args()

    print(f"Running {__file__} v{__version__}")

    panpipe_labels={}
    panpipe_labels = updateParams(panpipe_labels,'PWD',str(os.getcwd()))

    panpipeconfig_file=str(args.config_file)
    panpipeconfig_json=None
    if os.path.exists(panpipeconfig_file):
        with open(panpipeconfig_file, 'r') as infile:
            panpipeconfig_json = json.load(infile)
    panpipe_labels = process_labels(panpipeconfig_json,panpipeconfig_file,panpipe_labels,uselabel=False)

    pipeline=args.pipeline

    pipeline_class = getParams(panpipe_labels,"PIPELINE_CLASS")
    if pipeline_class is None:
        pipeline_class = pipeline

    use_pipeline_desc = getParams(panpipe_labels,"USE_PIPELINE_DESC")
    if use_pipeline_desc is None:
        use_pipeline_desc = ""

    pipeline_desc = getParams(panpipe_labels,"PIPELINE_DESC")
    if pipeline_desc is None or use_pipeline_desc == "N":
        pipeline_desc = pipeline 
    else:
        pipeline_desc = "".join([x if x.isalnum() else "_" for x in pipeline_desc]) 

    pipeline_outdir=os.path.join(getParams(panpipe_labels,"PIPELINE_DIR"),pipeline_desc)
    if not os.path.exists(pipeline_outdir):
        os.makedirs(pipeline_outdir,exist_ok=True)

    credentials = os.path.abspath(args.credentials)
    if os.path.exists(credentials):
        with open(credentials, 'r') as infile:
            cred_dict = json.load(infile)
            cred_user = getParams(cred_dict,"user")
            cred_password = getParams(cred_dict,"password")

    execution_json=getParams(panpipe_labels,"NIPYPE_CONFIG")
    if execution_json is None:
        execution_json = {} 

    participants_file = getParams(panpipe_labels,"PARTICIPANTS_FILE")
    bids_dir = getParams(panpipe_labels,"BIDS_DIR")

    participant_labels = getParams(panpipe_labels,"GROUP_PARTICIPANTS_LABEL")
    xnat_projects = getParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_PROJECT")
    session_labels = getParams(panpipe_labels,"GROUP_SESSION_LABEL")
    
    runGroupSubjects(participant_labels,xnat_projects, session_labels, pipeline=pipeline, pipeline_class=pipeline_class, pipeline_outdir=pipeline_outdir, panpipe_labels=panpipe_labels,bids_dir=bids_dir,cred_user=cred_user,cred_password=cred_password, execution_json=execution_json,analysis_level="group")



# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()


from panpipelines.pipelines import *
from panpipelines.scripts import *
import glob
import logging
from panpipelines.utils.util_functions import logger_setup

LOGGER = logger_setup("panpipelines.single_subject", logging.DEBUG)

class PANFactory:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(PANFactory, cls).__new__(cls)
            cls._instance.initialize()
        return cls._instance

    def initialize(self):
        # Initialization logic for the factory (e.g., setting up resources)
        LOGGER.debug("Initializing PANFactory.")
        root_module = "panpipelines"
        self.pipeline_module = f"{root_module}.pipelines"
        self.script_module = f"{root_module}.scripts"
        self.workflow_module = f"{root_module}.workflows"
        self.node_module = f"{root_module}.nodes"
        self.atlas_module = f"{root_module}.atlases"
        self.transform_module = f"{root_module}.transforms"

        LOGGER.debug("PANFactory ready. Modules defined.")
        LOGGER.debug(f"Pipelines: {self.pipeline_module}")
        LOGGER.debug(f"Scripts: {self.script_module}")
        LOGGER.debug(f"Workflows: {self.workflow_module}")
        LOGGER.debug(f"Nodes: {self.node_module}")
        LOGGER.debug(f"Transforms: {self.transform_module}")
        LOGGER.debug(f"Atlases: {self.atlas_module}")

    def get_node(self, name):

        try:
            # currently node and worfklow not defined as classes
            module = __import__(f"{self.node_module}.{name}",fromlist=[name])
            if module:
                return module
            else:
                return None
        except Exception as ex:
            pass

    def get_workflow(self, name):

        try:
            # currently node and worfklow not defined as classes
            module = __import__(f"{self.workflow_module}.{name}",fromlist=[name])
            if module:
                return module
            else:
                return None
        except Exception as ex:
            pass


    def get_atlas(self, name):
        try:
            module = __import__(f"{self.atlas_module}.{name}",fromlist=[name])
            if hasattr(module,name):
                atlasclass = getattr(module,name)
                return atlasclass
            else:
                return None
        except Exception as ex:
            pass

    def get_transform(self, name):
        try:
            module = __import__(f"{self.transform_module}.{name}",fromlist=[name])
            if hasattr(module,name):
                transformclass = getattr(module,name)
                return transformclass
            else:
                return None
        except Exception as ex:
            pass


    def get_pipeline(self, name):

        try:
            module = __import__(f"{self.pipeline_module}.{name}",fromlist=[name])
            if hasattr(module,name):
                pipeclass = getattr(module,name)
                return pipeclass
            else:
                raise None
        except Exception as ex:
            pass

    def get_script(self, name):
        try:
            module = __import__(f"{self.script_module}.{name}",fromlist=[name])
            if hasattr(module,name):
                scriptclass = getattr(module,name)
                return scriptclass
            else:
                return None
        except Exception as ex:
            pass

    def get_processflow(self, name):
        if "script" in name:
            processflow = self.get_script(name)
        elif "pipeline" in name:
            processflow = self.get_pipeline(name)           
        elif "workflow" in name:
            processflow = self.get_workflow(name)
        else:
            processflow = self.get_node(name)

        if not processflow:
            LOGGER.error(f"PAN object {name} not defined.")
            raise ValueError(f"{name} cannot be resolved as a PAN object by Factory")

        LOGGER.debug(f"PANfactory retrieving {processflow}")
        return processflow

    def get_PANclass(self, name):
        if "atlas" in name:
            panclass = self.get_atlas(name)
        elif "transform" in name:
            panclass = self.get_transform(name)
        else:
            panclass = self.get_processflow(name)

        if not panclass:
            raise ValueError(name)

        LOGGER.debug(f"PANfactory retrieving {panclass}")
        return panclass


    def search_PANclass(self, name):

        panclass = self.get_transform(name)

        if not panclass:
            panclass = self.get_atlas(name)

        if not panclass:
            panclass = self.get_pipeline(name)

        if not panclass:
            panclass = self.get_script(name)

        if not panclass:
            panclass = self.get_node(name)

        if not panclass:
            panclass = self.get_workflow(name)

        if not panclass:
            raise ValueError(name)

        LOGGER.debug(f"PANfactory retrieving {panclass}")
        return panclass

def getPANFactory():
    return PANFactory()


from nipype import logging as nlogging
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from pathlib import Path
from functools import partial
import os
import mne

IFLOGGER=nlogging.getLogger('nipype.interface')

def parse_params():
    parser = ArgumentParser(description="MNE make surface")
    parser.add_argument("--subject", help="Subject")
    parser.add_argument("--subjects_dir", type=Path, help="Freesurfer subjects directory")
    return parser

def make_surfaces(subject, subjects_dir):
    mne.bem.make_scalp_surfaces(subject, subjects_dir, force=True, overwrite=True, no_decimate=False, verbose=None)
    mne.bem.make_watershed_bem(subject, subjects_dir, overwrite=True, show=False)

def main():
    parser=parse_params()
    args, unknown_args = parser.parse_known_args()
    subject = args.subject
    subjects_dir = str(args.subjects_dir)

    make_surfaces(subject,subjects_dir)

# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging
import json
import sys
import nibabel
from nibabel.affines import apply_affine 
import numpy as np
import numpy.linalg as npl
from scipy import ndimage 
from skimage import measure
import shutil
import panpipelines.scripts.pancontainer_panscript as pancontainer_script
import panpipelines.scripts.mne_make_surfaces as mne_make_surfaces
import mne
from panpipelines.nodes.antstransform import antstransform_proc

# simulate strel('disk',5);
disc5 = np.zeros((9,9,9))
disc5[2:7,0:,0:]=1
disc5[0,2:7,2:7]=1
disc5[-1,2:7,2:7]=1
disc5[1,1:-1,1:-1]=1
disc5[7,1:-1,1:-1]=1

# simulate strel('disk',2);
disc2 = np.zeros((5,5,5))
disc2[2,2,2]=1
disc2[4,4,4]=1
disc2[2,0:,0:]=1
disc2[1,1:4,1:4]=1
disc2[3,1:4,1:4]=1
# simulate strel('disk',3)
disc3 = np.ones((5,5,5))

streldict={
    "disc5" : disc5,
    "disc3" : disc3,
    "disc2" : disc2

}

FRAC_INT_THRESH="0.7000000000000002"
VERT_GRAD="0"
RINGTHRESH = 0.15
STRELCLOSE = disc3
PHASESHIFT=14
PHASEAXIS=1


IFLOGGER=nlogging.getLogger('nipype.interface')

def returnCandidates(p0,p1,p2):
    arrayList=[]
    max_x = np.max(np.array([p0[0],p1[0],p2[0]]))
    max_y = np.max(np.array([p0[1],p1[1],p2[1]]))
    max_z = np.max(np.array([p0[2],p1[2],p2[2]]))
    min_x = np.min(np.array([p0[0],p1[0],p2[0]]))
    min_y = np.min(np.array([p0[1],p1[1],p2[1]]))
    min_z = np.min(np.array([p0[2],p1[2],p2[2]]))
    for x in range(min_x,max_x+1):
        for y in range(min_y, max_y + 1):
            for z in range(min_z,max_z + 1):
                p = np.array([x,y,z])
                if validCandidate(p0,p1,p2,p,tol=0.15):
                    arrayList.append(p)
    return arrayList

#https://math.stackexchange.com/questions/2582202/does-a-3d-point-lie-on-a-triangular-plane
def validCandidate(p0,p1,p2,p,tol):
    n = np.cross(p1-p0,p2-p0)/npl.norm(np.cross(p1-p0,p2-p0))
    d = np.dot(n,p-p1)
    if np.abs(d) < tol:
        return True
    else:
        return False

def derive_asl_artefact_v1(aslfile,aslfile_brain,aslfile_brain_mask,workdir,outputdir,RINGTHRESH=RINGTHRESH,STRELCLOSE=STRELCLOSE,PHASESHIFT=PHASESHIFT,PHASEAXIS=PHASEAXIS):

    data1 = nibabel.load(aslfile).get_fdata()
    data1 = data1[:,:,:,0]
    data2img = nibabel.load(aslfile_brain)
    data2 = data2img.get_fdata()
    data3img = nibabel.load(aslfile_brain_mask)
    data3 = data3img.get_fdata()

    diff12 = np.abs(data1-data2)
    diff12_img = nibabel.Nifti1Image(diff12,data2img.affine,data2img.header)
    ring = newfile(workdir,assocfile=aslfile,suffix="ring")
    nibabel.save(diff12_img,ring)

    diff12bw = diff12.copy()
    diff12bw_mask = diff12[:,:,:] > RINGTHRESH*np.max(diff12)
    diff12bw[diff12bw_mask]=1
    diff12bw[~diff12bw_mask]=0
    diff12bw = np.int16(diff12bw)
    diff12bw_shifted = np.roll(diff12bw,PHASESHIFT,PHASEAXIS)
    diff12bw_shifted_img = nibabel.Nifti1Image(diff12bw_shifted,data3img.affine,data3img.header)
    ring_shifted = newfile(workdir,assocfile=aslfile,suffix="ring_shifted")
    nibabel.save(diff12bw_shifted_img,ring_shifted)

    overlap_voxels = diff12bw_shifted + data3
    overlap_voxels_mask = overlap_voxels[:,:,:] == 2
    overlap_voxels[overlap_voxels_mask] = 1
    overlap_voxels[~overlap_voxels_mask] = 0
    overlap_voxels = np.int16(overlap_voxels)
    overlap_img = nibabel.Nifti1Image(overlap_voxels,data3img.affine,data3img.header)
    overlap = newfile(workdir,assocfile=aslfile,suffix="artefact_premorph")
    nibabel.save(overlap_img,overlap)

    overlap_voxels_morpho = ndimage.binary_closing(overlap_voxels, structure=STRELCLOSE).astype(np.int16)

    labels = measure.label(overlap_voxels_morpho)
    props = measure.regionprops(labels)
    largest_prop = max(props, key=lambda prop: prop.area)
    overlap_voxels_morpho_clean=np.zeros(overlap_voxels_morpho.shape)
    overlap_voxels_morpho_clean[tuple(largest_prop.coords.T)] = 1
    overlap_voxels_morpho_clean = np.int16(overlap_voxels_morpho_clean)
    overlap_clean_img = nibabel.Nifti1Image(overlap_voxels_morpho_clean,data3img.affine,data3img.header)
    overlap_clean = newfile(workdir,assocfile=aslfile,suffix="artefact")
    nibabel.save(overlap_clean_img,overlap_clean)

    final_clean = newfile(outputdir,assocfile=overlap_clean)
    shutil.copyfile(overlap_clean,final_clean)


def preproc_proc(labels_dict,bids_dir=""):

    cwd=os.getcwd()
    output_dir = cwd

    work_dir=os.path.join(output_dir,"preproc_workdir")
    if not os.path.exists(work_dir):
        os.makedirs(work_dir)

    labels_dict = updateParams(labels_dict,"CWD",cwd)

    command_base, container = getContainer(labels_dict,nodename="preproc", SPECIFIC="FSL_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the fsl version:")
    command = f"{command_base} fslversion"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    # set up dwi to process just the specific dwi session
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    subject = f"sub-{participant_label}"
    participant_session = getParams(labels_dict,'PARTICIPANT_SESSION')
    session = f"sub-{participant_session}"
    preproc_version = getParams(labels_dict,'PREPROC_VERSION')
    if preproc_version:
        VERSION_TO_RUN = preproc_version
    else:
        VERSION_TO_RUN = "2"

    # calculate asl artefact for acq-prod
    layout = BIDSLayout(bids_dir)
    asl=layout.get(subject=participant_label,session=participant_session,acquisition="prod",suffix='asl', extension='nii.gz')

    if len(asl) > 0:

        artefact_outputdir = getParams(labels_dict,'DERIVATIVES_DIR')
        if not artefact_outputdir:
            artefact_outputdir = os.path.join(output_dir,"asl_artefact", subject,session)
        else:
            artefact_outputdir = os.path.join(artefact_outputdir,"asl_artefact", subject,session)

        if not os.path.exists(artefact_outputdir):
            os.makedirs(artefact_outputdir, exist_ok=True)

        if VERSION_TO_RUN == "2":
            subjects_dir = getParams(labels_dict,'SUBJECTS_DIR')
            params =  f" --subject {subject}" \
                    f" --subjects_dir {subjects_dir}" 
            PREPROC_CONTAINER_TO_USE = getParams(labels_dict,"PREPROC_CONTAINER_TO_USE")
            if PREPROC_CONTAINER_TO_USE:
                panscript = pancontainer_script.pancontainer_panscript(labels_dict,params=params,command=f"python {mne_make_surfaces.__file__}",container_img=PREPROC_CONTAINER_TO_USE)
            else: 
                panscript = pancontainer_script.pancontainer_panscript(labels_dict,params=params,command=f"python {mne_make_surfaces.__file__}")
            panscript.run()

            os.chdir(output_dir)

            outer_skull=os.path.join(subjects_dir,subject,'bem','outer_skull.surf')
            rr_mm_outer_skull,tris_outer_skull = mne.read_surface(outer_skull)

            orig_file=f"{subjects_dir}/{subject}/mri/T1.mgz"
            origimg = nibabel.load(orig_file)
            Torig = origimg.header.get_vox2ras_tkr()
            origimg_dtype = origimg.header.get_data_dtype()
            origimg_shape = origimg.shape
            outer_skull_vol = np.zeros((origimg_shape),dtype=origimg_dtype)
            inv_Torig = npl.inv(Torig)

            for rr in rr_mm_outer_skull:
                vox = apply_affine(inv_Torig, rr)
                vox_ind = tuple(np.round(vox).astype(int))
                outer_skull_vol[vox_ind] = 1

            for tri in tris_outer_skull:
                p0 = np.round(apply_affine(inv_Torig,rr_mm_outer_skull[tri[0]])).astype(int)
                p1 = np.round(apply_affine(inv_Torig,rr_mm_outer_skull[tri[1]])).astype(int)
                p2 = np.round(apply_affine(inv_Torig,rr_mm_outer_skull[tri[2]])).astype(int)
                valid_vox = returnCandidates(p0,p1,p2)
                for vox in valid_vox:
                    vox_ind = tuple(vox.astype(int))
                    outer_skull_vol[vox_ind] = 1
            outer_skull_img = nibabel.Nifti1Image(outer_skull_vol,origimg.affine,origimg.header)
            outer_skull_img_file=os.path.join(work_dir,f"{subject}_{session}_outer_skull.nii.gz")
            nibabel.save(outer_skull_img,outer_skull_img_file) 

            transform_list =  getParams(labels_dict,'TRANSFORM_MAT')
            transform_ref =  getParams(labels_dict,'TRANSFORM_REF')
            outer_skull_aslspace_file = None
            if transform_list and transform_ref:
                results = antstransform_proc(labels_dict, outer_skull_img_file,transform_list, transform_ref)
                outer_skull_aslspace_file = results['out_file']     

            outer_skull_aslspace_bin=newfile(work_dir,outer_skull_aslspace_file,suffix="bin")
            command = f"{command_base} fslmaths {outer_skull_aslspace_file} -bin {outer_skull_aslspace_bin}"
            evaluated_command=substitute_labels(command, labels_dict)
            results = runCommand(evaluated_command,IFLOGGER)

            outskull_aslspace_img=nibabel.load(outer_skull_aslspace_bin)
            outskull_aslspace_data = outskull_aslspace_img.get_fdata()
            outskull_aslspace_data_close = ndimage.binary_closing(outskull_aslspace_data, structure=disc5).astype(np.int16)
            outskull_aslspace_data_close_file=newfile(work_dir,outer_skull_aslspace_bin,suffix="close-disc-5")
            outskull_aslspace_data_close_img = nibabel.Nifti1Image(outskull_aslspace_data_close,outskull_aslspace_img.affine,outskull_aslspace_img.header)
            nibabel.save(outskull_aslspace_data_close_img,outskull_aslspace_data_close_file)  

            outer_skull_shift=newfile(work_dir,outskull_aslspace_data_close_file,suffix="shift")
            data_shifted = np.roll(outskull_aslspace_data_close,PHASESHIFT,1)
            data_shifted[:,0:PHASESHIFT,:]=0
            shifted_img = nibabel.Nifti1Image(data_shifted,outskull_aslspace_img.affine,outskull_aslspace_img.header)
            nibabel.save(shifted_img,outer_skull_shift)

            outer_skull_shift_final = newfile(artefact_outputdir,assocfile=f"{subject}_{session}_asl_chemical_shift_artefact.nii.gz")
            shutil.copyfile(outer_skull_shift,outer_skull_shift_final)  

        else:
            FRAC_INT_THRESH_lookup= getParams(labels_dict,'FRAC_INT_THRESH')
            if FRAC_INT_THRESH_lookup:
                FRAC_INT_THRESH = FRAC_INT_THRESH_lookup

            VERT_GRAD_lookup=getParams(labels_dict,'VERT_GRAD')
            if VERT_GRAD_lookup:
                VERT_GRAD = VERT_GRAD_lookup

            RINGTHRESH_lookup=getParams(labels_dict,'RINGTHRESH')
            if RINGTHRESH_lookup:
                RINGTHRESH = np.float64(RINGTHRESH_lookup)

            STRELCLOSE_lookup=getParams(labels_dict,'STRELCLOSE')
            if STRELCLOSE_lookup and STRELCLOSE_lookup in streldict.keys():
                STRELCLOSE = streldict[STRELCLOSE_lookup]

            asl_bidsfile=asl[0]
            asl_file=asl_bidsfile.path
            asl_file_brain = newfile(work_dir,assocfile=asl_file,suffix="brain")
            asl_file_brain_mask = newfile(work_dir,assocfile=asl_file_brain,suffix="mask")

            params=f" {asl_file}" \
                f" {asl_file_brain}" \
                " -R" \
                f" -f {FRAC_INT_THRESH}" \
                f" -g {VERT_GRAD}" \
                " -m -s"

            command=f"{command_base} bet"\
                " "+params
            evaluated_command=substitute_labels(command, labels_dict)
            results = runCommand(evaluated_command,IFLOGGER)

            derive_asl_artefact_v1(asl_file,asl_file_brain,asl_file_brain_mask,work_dir,artefact_outputdir,RINGTHRESH=RINGTHRESH,STRELCLOSE=STRELCLOSE)

            asl_entities = asl_bidsfile.get_entities()
            m0_entities = asl_entities.copy()
            m0_entities["suffix"]="m0scan"
            m0  = layout.get(return_type='file', invalid_filters='allow', **m0_entities)
            if len(m0) > 0:
                m0_file=m0[0]
                m0_file_brain = newfile(work_dir,assocfile=m0_file,suffix="brain")
                m0_file_brain_mask = newfile(work_dir,assocfile=m0_file_brain,suffix="mask")

                params=f" {m0_file}" \
                    f" {m0_file_brain}" \
                    " -R" \
                    f" -f {FRAC_INT_THRESH}" \
                    f" -g {VERT_GRAD}" \
                    " -m -s"

                command=f"{command_base} bet"\
                    " "+params
                evaluated_command=substitute_labels(command, labels_dict)
                results = runCommand(evaluated_command,IFLOGGER)

                derive_asl_artefact_v1(m0_file,m0_file_brain,m0_file_brain_mask,work_dir,artefact_outputdir,RINGTHRESH=RINGTHRESH,STRELCLOSE=STRELCLOSE)


    return {
        "output_dir":output_dir
    }



class preprocInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class preprocOutputSpec(TraitedSpec):
    output_dir = traits.String(desc="QSIPREP output directory")
    
    
class preproc_pan(BaseInterface):
    input_spec = preprocInputSpec
    output_spec = preprocOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = preproc_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='preproc_node',bids_dir="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(preproc_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import numpy as np 
import nibabel as nib
import pandas as pd
import json
import datetime
from pathlib import Path
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def roi_mean_single_proc(labels_dict,input_file,atlas_file,atlas_index):

    metadata_comments=""
    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    output_dir=cwd
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')

    command_base, container = getContainer(labels_dict,nodename="roi_mean_single", SPECIFIC="FSL_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the fsl version:")
    command = f"{command_base} fslversion"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    if not session_label:
        roi_output_dir = os.path.join(cwd,f"sub-{participant_label}_roi_output_dir")
    else:
        roi_output_dir = os.path.join(cwd,f"sub-{participant_label}_ses-{session_label}_roi_output_dir")

    if not os.path.isdir(roi_output_dir):
        os.makedirs(roi_output_dir)

    if Path(atlas_file).suffix == ".mgz":
        mgzdir = os.path.join(cwd,'mgz_nii')
        if not os.path.isdir(mgzdir):
            os.makedirs(mgzdir)

        fs_command_base, fscontainer = getContainer(labels_dict,nodename="convMGZ2NII",SPECIFIC="FREESURFER_CONTAINER",LOGGER=IFLOGGER)
        atlas_file_nii = newfile(mgzdir,atlas_file,extension=".nii.gz")
        convMGZ2NII(atlas_file, atlas_file_nii, fs_command_base)
        atlas_file = atlas_file_nii

    if Path(input_file).suffix == ".mgz":
        mgzdir = os.path.join(cwd,'mgz_nii')
        if not os.path.isdir(mgzdir):
            os.makedirs(mgzdir)

        fs_command_base, fscontainer = getContainer(labels_dict,nodename="convMGZ2NII",SPECIFIC="FREESURFER_CONTAINER",LOGGER=IFLOGGER)
        input_file_nii = newfile(mgzdir,input_file,extension=".nii.gz")
        convMGZ2NII(input_file, input_file_nii, fs_command_base)
        input_file = input_file_nii

    if not session_label:
        roi_raw_txt = os.path.join(roi_output_dir,f"{participant_label}_roi_raw.txt")
    else:
        roi_raw_txt = os.path.join(roi_output_dir,f"{participant_label}_{session_label}_roi_raw.txt")


    atlas_type="3D"

    atlas_img  = nib.load(atlas_file)
    image_shape = atlas_img.header.get_data_shape()
    if len(image_shape) > 3:
        atlas_type="4D"

    if atlas_type == "4D":
        params = " -i "+input_file+ \
            " -d "+atlas_file + \
            " -o "+roi_raw_txt

        command=f"{command_base} fsl_glm"\
                " "+params

        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)
    else:
        params = " -i "+input_file+ \
            " -o "+roi_raw_txt+\
            " --label="+atlas_file
            
        command=f"{command_base} fslmeants"\
                " "+params

        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

    atlas_index_mode = None
    if getParams(labels_dict,'ATLAS_INDEX_MODE'):
        atlas_index_mode = getParams(labels_dict,'ATLAS_INDEX_MODE')
    elif getParams(labels_dict,'NEWATLAS_INDEX_MODE'):
        atlas_index_mode = getParams(labels_dict,'NEWATLAS_INDEX_MODE')

    if atlas_index.split(":")[0] == "get_freesurfer_atlas_index":
        lutfile = substitute_labels(atlas_index.split(":")[1],labels_dict)
        new_atlas_index=newfile(roi_output_dir,atlas_file,suffix="desc-index",extension=".txt")
        atlas_index_json = newfile(roi_output_dir,new_atlas_index,extension="json")
        atlas_dict,atlas_index_out=get_freesurferatlas_index_mode(atlas_file,lutfile,new_atlas_index,atlas_index_mode=atlas_index_mode)
        atlas_index=new_atlas_index
        export_labels(atlas_dict,atlas_index_json)

    with open(atlas_index, 'r') as in_file:
        lines = in_file.readlines()

    table_columns = [x.replace('\n','') for x in lines]
    df2 = pd.read_table(roi_raw_txt,sep=r"\s+",header=None)
    numrows = len(df2)

    # check that we are not missing ROIS
    if len(table_columns) > len(df2.columns):
        IFLOGGER.warn(f"Size of data columns in {roi_raw_txt} is less than expected. {len(df2.columns)} instead of {len(table_columns)}.")
        IFLOGGER.warn(f"It is possible that due to a transformation to lower resolution some smaller ROIs were lost.")
        IFLOGGER.warn(f"checking {atlas_file} for missing ROIs.")

        if metadata_comments:
            metadata_comments = metadata_comments + f"Size of data columns in {roi_raw_txt} is less than expected. {len(df2.columns)} instead of {len(table_columns)}."
        else:
            metadata_comments = f"Size of data columns in {roi_raw_txt} is less than expected. {len(df2.columns)} instead of {len(table_columns)}."

        atlasimg = nib.load(atlas_file)
        atlasimg_data = atlasimg.get_fdata()
        for roi_check in range(len(table_columns)):
            roi_sum = np.sum(atlasimg_data  == (roi_check + 1))
            if roi_sum == 0:
                IFLOGGER.warn(f"ROI Index {roi_check + 1} is missing. This corresponds to ROI Label {table_columns[roi_check]}")

                if metadata_comments:
                    metadata_comments = metadata_comments + f"Missing ROI >>  [Index {roi_check + 1} : {table_columns[roi_check]} ], "
                else:
                    metadata_comments = f"Missing ROI >>  [Index {roi_check + 1} : {table_columns[roi_check]} ]  "


    elif len(table_columns) < len(df2.columns):
        IFLOGGER.error(f"Size of data columns in {roi_raw_txt} is larger than expected. {len(df2.columns)} instead of {len(table_columns)}.")
        IFLOGGER.error(f"This suggests that there may be a mismatch between the atlas index {atlas_index} uand the atlas {atlas_file}.")
        raise ValueError(f"Size of data columns in {roi_raw_txt} is larger than expected. {len(df2.columns)} instead of {len(table_columns)}\nThis suggests that there may be a mismatch between the atlas index {atlas_index} uand the atlas {atlas_file}.")


    
    csv_basename = ""

    atlas_name = getParams(labels_dict,'ATLAS_NAME')
    if not atlas_name:
        atlas_name = os.path.basename(atlas_file).split('.')[0].split('_')[0].split('-')[0]
    csv_basename = f"atlas-{atlas_name}"

    modality = get_bidstag("desc",input_file,True)
    if modality:
        modality=modality[-1].split(".")[0]
        csv_basename = csv_basename + "_" + modality
    else:
        modality = os.path.basename(input_file).split('.')[0].split('_')[-1].split('-')[-1]
        csv_basename = csv_basename + "_" + modality

    table_columns = [f"{atlas_name}.{x}.{modality}" for x in table_columns]

    if not session_label:
        csv_basename  = f"sub-{participant_label}" + "_" + csv_basename
    else:
        csv_basename  = f"sub-{participant_label}_ses-{session_label}" + "_" + csv_basename

    roi_csv = os.path.join(roi_output_dir,'{}.csv'.format(csv_basename))

    if numrows < 2:
        if session_label:
            df2.insert(0,"session_id",["ses-"+session_label])
            table_columns.insert(0,"session_id")
        else:
            df2.insert(0,"session_id",["NotProvided"])
            table_columns.insert(0,"session_id")

        df2.insert(0,"subject_id",["sub-"+participant_label])
        table_columns.insert(0,"subject_id")

        df2.columns = table_columns
        df2.to_csv(roi_csv,sep=",",header=True, index=False)
    else:
        flat_vals=[]
        flat_tablecolumns=[]
        for count in range(numrows):
            flat_vals.extend(df2.iloc[count].to_list())
            flat_tablecolumns.extend([x + f".{count}" for x in table_columns])
        newdf=pd.DataFrame([flat_vals])
        if session_label:
            newdf.insert(0,"session_id",["ses-"+session_label])
            flat_tablecolumns.insert(0,"session_id")
        else:
            newdf.insert(0,"session_id",["NotProvided"])
            flat_tablecolumns.insert(0,"session_id")
        newdf.insert(0,"subject_id",["sub-"+participant_label])
        flat_tablecolumns.insert(0,"subject_id")
        newdf.columns = flat_tablecolumns
        newdf.to_csv(roi_csv,sep=",",header=True, index=False)


    metadata = {}
    roi_json = os.path.join(roi_output_dir,'{}.json'.format(csv_basename))
    metadata = updateParams(metadata,"Title","roi_mean_single")
    metadata = updateParams(metadata,"Description","Extract Measures from Image file using provided atlas.")
    metadata = updateParams(metadata,"DateCreated",datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S:%f"))
    metadata = updateParams(metadata,"Atlas File",atlas_file)
    metadata = updateParams(metadata,"Atlas Labels",atlas_index)
    metadata = updateParams(metadata,"Input File",input_file)
    metadata = updateParams(metadata,"Command",evaluated_command)
    if metadata_comments:
        metadata = updateParams(metadata,"Comments",metadata_comments)

    export_labels(metadata,roi_json)


    out_files=[]
    out_files.insert(0,roi_csv)

    return {
        "roi_csv":roi_csv,
        "roi_output_dir":roi_output_dir,
        "output_dir":output_dir,
        "out_files":out_files
    }



class roi_mean_singleInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    input_file = File(desc="input file")
    atlas_file = File(desc='atlas file')
    atlas_index = File(desc='atlas index file')

class roi_mean_singleOutputSpec(TraitedSpec):
    roi_csv = File(desc='CSV file of results')
    roi_output_dir = traits.String(desc='roi output dir')
    output_dir = traits.String(desc='output dir')
    out_files = traits.List(desc='list of files')
    
class roi_mean_single_pan(BaseInterface):
    input_spec = roi_mean_singleInputSpec
    output_spec = roi_mean_singleOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = roi_mean_single_proc(
            self.inputs.labels_dict,
            self.inputs.input_file,
            self.inputs.atlas_file,
            self.inputs.atlas_index,
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="roi_mean_single_node",input_file="",atlas_file="",atlas_index="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(roi_mean_single_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    pan_node.inputs.input_file = input_file
    pan_node.inputs.atlas_file =  atlas_file
    pan_node.inputs.atlas_index =  atlas_index

    return pan_node


